// __________________________________________________________________________________________________
//    GLGooey Graphical User Interface for OpenGL
//    Copyright (c) 2004 Niel Waldren
//
// This software is provided 'as-is', without any express or implied warranty. In no event will
// the authors be held liable for any damages arising from the use of this software.
//
// Permission is granted to anyone to use this software for any purpose, including commercial
// applications, and to alter it and redistribute it freely, subject to the following restrictions:
//
//     1. The origin of this software must not be misrepresented; you must not claim that you
//        wrote the original software. If you use this software in a product, an acknowledgment
//        in the product documentation would be appreciated but is not required.
//
//     2. Altered source versions must be plainly marked as such, and must not be misrepresented
//        as being the original software.
//
//     3. This notice may not be removed or altered from any source distribution.
//
// __________________________________________________________________________________________________
#include "png.h"


#include "PNGImage.h"




//! The number of bytes used to check whether a file is actually a PNG file
const unsigned int numBytesForPNGCheck = 4;

// >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
bool
isValidPNGFile(FILE *fp)
{
    png_byte buf[numBytesForPNGCheck];

    for(size_t i = 0; i < numBytesForPNGCheck; i++)
    {
        if (!(buf[i] = fgetc(fp))) return false; // ### RETURN ###
    }
    return (!png_sig_cmp(buf, 0, numBytesForPNGCheck));
}
// <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<





// >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
#ifndef png_jmpbuf
#define png_jmpbuf(png_ptr) ((png_ptr)->jmpbuf)
#endif
// <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<




// >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
void
user_read_data(png_structp png_ptr, png_bytep data, png_size_t length)
{
    voidp read_io_ptr = png_get_io_ptr(png_ptr) ;
    fread( (unsigned char*) data, length, 1, (FILE*) read_io_ptr ) ;
}
// <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<




// >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
PNGImage::PNGImage(const std::string& aFileName) :
    data_(0)
{
	std::string localFilename = aFileName;
#ifdef _WINDOWS
	int it = localFilename.find('/');
	while(it != std::string::npos){
		localFilename = localFilename.substr(0,it)+"\\"+localFilename.substr(it+1);
		it = localFilename.find('/');
	}
#endif
    FILE *fp = fopen(aFileName.c_str(), "rb");
    if(fp == 0)
    {
        printf("File could not be opened.");
    }
    if(!isValidPNGFile(fp))
    {
        fclose(fp);
        printf("File is not a valid png file");
    }


    png_structp png_ptr = png_create_read_struct(PNG_LIBPNG_VER_STRING, 0, 0, 0);
    if(png_ptr == 0)
    {
        fclose(fp);
        printf("png_create_read_struct failed for file");
    }

    png_infop info_ptr = png_create_info_struct(png_ptr);
    if(info_ptr == 0)
    {
        png_destroy_read_struct(&png_ptr, png_infopp(0), png_infopp(0));
        fclose(fp);
        printf("png_create_info_struct failed for file");
    }


    if(setjmp(png_jmpbuf(png_ptr)))
    {
        png_destroy_read_struct(&png_ptr, png_infopp(0), png_infopp(0));
        fclose(fp);
        printf("setjmp failed for file");
    }

    png_set_read_fn( png_ptr, fp, (png_rw_ptr)user_read_data ) ;	

    png_init_io(png_ptr, fp);
    png_set_sig_bytes(png_ptr, numBytesForPNGCheck);

    png_read_png(png_ptr, info_ptr, PNG_TRANSFORM_IDENTITY, 0);
    const unsigned int width = png_get_image_width(png_ptr, info_ptr);
    const unsigned int height = png_get_image_height(png_ptr, info_ptr);
    const unsigned int bit_depth = png_get_bit_depth(png_ptr, info_ptr);
    if(bit_depth != 8)
    {
        fclose(fp);
        png_destroy_info_struct(png_ptr, png_infopp(&info_ptr));
        png_destroy_read_struct(&png_ptr, png_infopp(0), png_infopp(0));
        printf("File has an unsupported bit depth. The bit depth must be 8.");
    }
    const png_byte colorType = png_get_color_type(png_ptr, info_ptr);
    if((colorType != PNG_COLOR_TYPE_RGB) && (colorType != PNG_COLOR_TYPE_RGB_ALPHA))
    {
        fclose(fp);
        png_destroy_info_struct(png_ptr, png_infopp(&info_ptr));
        png_destroy_read_struct(&png_ptr, png_infopp(0), png_infopp(0));
        printf("File has an unsupported color type. RGB and RGBA images are supported.");
    }

    const int bytesPerPixel = (colorType == PNG_COLOR_TYPE_RGB) ? 3 : 4;
    const int stride = bytesPerPixel * width;

    unsigned char* data = new unsigned char[stride * height];
    png_bytep* row_pointers = png_get_rows(png_ptr, info_ptr);
    for(unsigned int i = 0; i < height; ++i)
    {
        const unsigned int row = height - i - 1;
        memcpy(data + (row * stride), row_pointers[i], stride);
    }
	width_ = width;
	height_ = height;
	
    numberOfBitsPerPixel_ = bytesPerPixel * 8;
    data_ = data;

    fclose(fp);
    png_destroy_info_struct(png_ptr, png_infopp(&info_ptr));
    png_destroy_read_struct(&png_ptr, png_infopp(0), png_infopp(0));
}
// <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<







// >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
PNGImage::~PNGImage()
{

	if(data_)
		delete[] data_;
}
// <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<

